var async = require('async'),
    debug = require('debug')('async-queue-improved'),
    EventEmitter = require('events').EventEmitter

module.exports = function(worker, concurrency) {
  var tasks = {}

  // task: {id: 1, data: 'any', timeout: 12311}
  var queue = async.queue(function(task, callback) {
    debug("Got task %s", task.id)

    var retryCallback = function(err) {
      if (!err) return callback.apply(null, arguments)
      if (!err.timeout) return callback.apply(null, arguments)

      if (!task.retries) {
        // store the error
        queue.errors.push({ err: err, task: task })
        return callback(err)
      }

      // console.error(err.stack || err)
      task.attempts = task.attempts || 0
      task.attempts++

      if (task.attempts <= task.retries) {
        queue.push(task, callback)
      }
    }

    // Set timeout for the task execution
    var timeout = setTimeout(function() {
      timeout = null
      debug("Task %s is timed out", task.id)

      var err = new Error("Task " + task.id + " has timed out (timeout was " + task.timeout + "ms)")
      err.timeout = true
      // pass the error to a callback
      retryCallback(err)
    }, task.timeout)

    // Wrapping callback to respect timeout
    var callbackCalled = false
    var callbackTimeout = function(err, result) {
      if (callbackCalled) {
        debug("Task %s callback called more than once!", task.id)
        return
      }
      callbackCalled = true

      // Check if we haven't timed out
      if (timeout) {
        clearTimeout(timeout)
        setImmediate(function(){
          retryCallback(err, result)
        })
      }
    }

    // If the task is already registered
    if (tasks[task.id]) {
      debug("Task %s is already registered", task.id)

      // If the task is already done, call the callback with result
      if (tasks[task.id].resultArgs) {
        debug("Task %s is already done, giving the result", task.id)

        callbackTimeout.apply(null, tasks[task.id].resultArgs)
      }
      // Subscribe on task complete
      else {
        debug("Task %s is in progress, subscribe on 'complete' event", task.id)

        tasks[task.id].emitter.once('complete', function(resultArgs) {
          callbackTimeout.apply(null, resultArgs)
        })
      }
      return
    }

    // If we are still here, it means that nobody has runned the task yet
    // so we will run it
    debug("Task %s is new, start working on it", task.id)

    // callback wich emits events to task emitter
    var emitterCallback = function(err, result) {
      debug("Task %s complete, took %dms, %d listeners",
          task.id, new Date - tasks[task.id].startedAt, (tasks[task.id].emitter._events['complete'] || []).length + 1)

      // store the error
      if (err) {
        queue.errors.push({ err: err, task: task })
      }

      // remember the result and emit 'complete' event with result argument
      tasks[task.id].resultArgs = [err, result]
      tasks[task.id].emitter.emit('complete', tasks[task.id].resultArgs)
      // call the callback for original task
      callbackTimeout(err, result)
    }

    // Register task
    tasks[task.id] = {
      emitter: new EventEmitter,
      resultArgs: null,
      startedAt: new Date
    }

    tasks[task.id].emitter.setMaxListeners(100)

    // Run worker on this task
    worker(task.data, emitterCallback)
  }, concurrency)

  var finish = function(callback) {
    var errors = queue.errors.length > 0 ? queue.errors : null
    callback(errors)
    queue.drain = null
  }
  
  // Convenience onComplete function
  queue.onComplete = function(callback) {
    if (queue.length() + queue.running() > 0) {
      queue.drain = finish.bind(null, callback)
    }
    else {
      finish(callback)
    }
  }

  // Collect errors inside the queue
  queue.errors = []

  return queue
}
